@@ -6,6 +6,7 @@ from typing import List
 import mlcf.datatools.indice_tools as i_tools
 import numpy as np
 import pandas as pd
+import pandas_ta as _  # noqa
 import talib.abstract as ta
 
 
@@ -151,6 +152,9 @@ class Indice(Enum):
     # volatility
     VOLATILITY = "VOLATILITY"
 
+    # return
+    RETURN = "RETURN"
+
     @classmethod
     def list_value(self):
         return [item.value for item in list(self)]
@@ -378,16 +382,19 @@ def add_indicator(data: pd.DataFrame, indice_name: Indice):
         ).shift(-26)
 
     elif case(Indice.SMA1):
-        dataframe["SMA1"] = add_SMA1(dataframe)
+        dataframe = add_SMA1(dataframe)
 
     elif case(Indice.VOLATILITY):
-        dataframe["volatility"] = add_volatility(dataframe)
+        dataframe = add_volatility(dataframe)
 
     elif case(Indice.LNSMA1):
-        dataframe["LNSMA1"] = add_ln_SMA1(dataframe)
+        dataframe = add_ln_SMA1(dataframe)
 
     elif case(Indice.PERCENTGROWTH):
-        dataframe["percent_growth"] = add_percent_growth(dataframe)
+        dataframe = add_percent_growth(dataframe)
+
+    elif case(Indice.RETURN):
+        dataframe = add_return(dataframe)
 
     else:
         raise Exception("Unknown indice")
@@ -471,3 +478,10 @@ def add_volatility(data: pd.DataFrame):
             if ('growth'+str(offset)) in data:
                 dataframe.drop(('growth'+str(offset)), axis=1, inplace=True)
     return dataframe
+
+
+def add_return(data: pd.DataFrame, offset: int = 1, colname: str = "return", dropna: bool = False):
+    dataframe = data.copy()
+    dataframe[colname] = np.log(dataframe.close) - np.log(dataframe.shift(offset).close)
+    dataframe.dropna(inplace=dropna)
+    return dataframe
